import unittest
import io

from thdl.generate.enumeration import Enumeration

SINGLE_LINE = "type t_my_enum is (ONE, TWO, THREE);"

SINGLE_GOLDEN_PKG = """function to_slv (e : t_my_enum) return std_logic_vector;

function to_my_enum (slv : std_logic_vector(1 downto 0)) return t_my_enum;
"""

SINGLE_GOLDEN_PKG_BODY = """function to_slv (e : t_my_enum) return std_logic_vector is
begin
   case e is
      when ONE => return "00";
      when TWO => return "01";
      when THREE => return "10";
   end case;
end function;

function to_my_enum (slv : std_logic_vector(1 downto 0)) return t_my_enum is
begin
   case slv is
      when "00" => return ONE;
      when "01" => return TWO;
      when "10" => return THREE;
      when others => return THREE;
   end case;
end function;
"""


MULTI_LINE = """type my_enum is (
    foo,
    bar,
    zaz
);
"""

MULTI_GOLDEN_PKG = """function to_slv (e : my_enum) return std_logic_vector;

function to_my_enum (slv : std_logic_vector(1 downto 0)) return my_enum;
"""

MULTI_GOLDEN_PKG_BODY = """function to_slv (e : my_enum) return std_logic_vector is
begin
   case e is
      when foo => return "00";
      when bar => return "01";
      when zaz => return "10";
   end case;
end function;

function to_my_enum (slv : std_logic_vector(1 downto 0)) return my_enum is
begin
   case slv is
      when "00" => return foo;
      when "01" => return bar;
      when "10" => return zaz;
      when others => return zaz;
   end case;
end function;
"""


class TestEnumeration(unittest.TestCase):
    def test_single_line(self):
        fh = io.StringIO(SINGLE_LINE)

        for l in fh:
            line = l
            break

        enum = Enumeration(fh, line)

        pkg = enum.generate_pkg()
        self.assertEqual(pkg, SINGLE_GOLDEN_PKG)

        pkg_body = enum.generate_pkg_body()
        self.assertEqual(pkg_body, SINGLE_GOLDEN_PKG_BODY)

    def test_multi_line(self):
        fh = io.StringIO(MULTI_LINE)

        for l in fh:
            line = l
            break

        enum = Enumeration(fh, line)

        pkg = enum.generate_pkg()
        self.assertEqual(pkg, MULTI_GOLDEN_PKG)

        pkg_body = enum.generate_pkg_body()
        self.maxDiff = None
        self.assertEqual(pkg_body, MULTI_GOLDEN_PKG_BODY)



import unittest
import io

from thdl.check.reset import check

LINES_STUCK_TO_1 = """rst_p => '1',
rst_p=>'1',
reset_p_i=>'1',
rst => '1',
reset => '1',
reset_i => '1',
reset_p_i => '1',
RST_P_I=>'1');
wb_rst_p=> '1',
wb_rst_p_i=> '1',
foo_bar_reset => '1',
"""

LINES_MAPPED_TO_NEGATIVE = """rst_p => rstn,
rstp => rstn,
reset => reset_n_i,
reset_p_i => rst_n);
wb_rst_p=>  rst_n,
wb_rst_p_i=> foo_reset_n,
"""

LINES_NEGATED_POSITIVE ="""rst_p => not rst_p_i,
reset => not(rst_p),
rst_i => not wb_resetp,
"""

LINES_VALID ="""rst_p => '0',
rst_p_i => not rst_n_i,
rst_p_i => not(reset_n),
reset => not wb_rstn_i
rst_i => not wb_resetn,
"""

class TestPositiveReset(unittest.TestCase):
    def test_stuck_to_1(self):
        fh = io.StringIO(LINES_STUCK_TO_1)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, "Positive reset stuck to '1'!", l)

    def test_mapped_to_negative(self):
        fh = io.StringIO(LINES_MAPPED_TO_NEGATIVE)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, "Positive reset mapped to negative reset!", l)

    def test_negated_positive(self):
        fh = io.StringIO(LINES_NEGATED_POSITIVE)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, "Positive reset mapped to negated positive reset!", l)

    def test_valid(self):
        fh = io.StringIO(LINES_VALID)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, None, l)



from . import reset

from . import file_info


def check(filepaths):
    num_violations = 0

    for path in filepaths:
        file_info.FILEPATH = path
        with open(path, encoding='latin-1') as fh:
            for i, line in enumerate(fh, start=1):
                file_info.LINE = line
                file_info.LINE_NUMBER = i

                # Preprocess line.
                l = line.lower().strip()
                if l.startswith("--"):
                    continue
                l = line.split("--")[0]

                if reset.check(l):
                    num_violations += 1

    print("Checked %d files." % len(filepaths))
    if num_violations > 0:
        if num_violations == 1:
            print(f"Found 1 violation.")
        else:
            print(f"Found {num_violations} violations.")
        return 1

    print(f"Found 0 violations.")
    return 0



import math
import re


class Enumeration:
    def __init__(self, fh, line):
        line = line.split("--")[0].strip()

        self.identifier = line.split()[1]

        self.values = []
        # Single line declaration
        match = re.match("type\s+\w+\s+is\s*\((.*)\);", line)
        if match:
            self.values = list(map(lambda v: v.strip(), match.group(1).split(",")))
        elif line.endswith("("):
            for line in fh:
                if re.match("^\s*\)\s*;", line):
                    break

                line = line.split("--")[0].strip()

                values = line.split(",")
                for v in values:
                    if v:
                        self.values.append(v.strip())
                if re.match("\)\s*;", line):
                    break
        else:
            raise Exception(f"Unable to parse enumeration declaration: {line}")

        self.length = math.ceil(math.log2(len(self.values)))

    def generate_pkg(self):
        string = (
            "function to_slv (e : " + self.identifier + ") return std_logic_vector;\n\n"
        )

        if self.identifier.startswith("t_") or self.identifier.startswith("T_"):
            tmp = self.identifier[2:]
        else:
            tmp = self.identifier
        name = "to_" + tmp
        string += (
            "function "
            + name
            + " (slv : std_logic_vector("
            + str(self.length - 1)
            + " downto 0)) return "
            + self.identifier
            + ";\n"
        )

        return string

    def generate_pkg_body(self):
        string = (
            "function to_slv (e : " + self.identifier + ") return std_logic_vector is\n"
        )
        string += "begin\n"
        string += "   case e is\n"
        for i, v in enumerate(self.values):
            string += (
                "      when "
                + v
                + ' => return "'
                + bin(i)[2:].zfill(self.length)
                + '";\n'
            )
        string += "   end case;\n"
        string += "end function;\n\n"

        if self.identifier.startswith("t_") or self.identifier.startswith("T_"):
            tmp = self.identifier[2:]
        else:
            tmp = self.identifier
        name = "to_" + tmp
        string += (
            "function "
            + name
            + " (slv : std_logic_vector("
            + str(self.length - 1)
            + " downto 0)) return "
            + self.identifier
            + " is\n"
        )
        string += "begin\n"
        string += "   case slv is\n"
        for i, v in enumerate(self.values):
            string += (
                '      when "'
                + bin(i)[2:].zfill(self.length)
                + '" => return '
                + v
                + ";\n"
            )
        # TODO: Think about more elegant others handling. It should probably assert.
        string += "      when others => return " + self.values[-1] + ";\n"
        string += "   end case;\n"
        string += "end function;\n"

        return string



"""
SPDX-License-Identifier: GPL-3.0-or-later

Copyright (c) 2021 Michał Kruszewski
"""

import argparse
import pathlib
import os
import sys

from .check import check
from .generate import generate


VERSION = "0.0.6"


def parse_cmd_line_args():
    parser = argparse.ArgumentParser(
        prog="thdl",
        description="thdl is a bunch of tools (hidden under single program) for easing the work with VHDL language."
        " It is (and will always be) based solely on the text processing, no semantic analysis."
        " Such approach draws clear line what might be included and what will never be supported."
        " 't' in the 'thdl' stands for 'text'.",
    )

    parser.add_argument(
        "-v", "--version", help="Display thdl version.", action="version", version=VERSION
    )

    subparsers = parser.add_subparsers()

    path_help = "Path to the file or directory."
    "In case of the directory the processing is done for all files and subdirectories in a recursive way."

    check_parser = subparsers.add_parser(
        "check",
        help="Check for extremely dump mistakes such as stucking resets to constant reset value.",
    )
    check_parser.add_argument("path", help=path_help)
    check_parser.set_defaults(func=check.check)

    generate_parser = subparsers.add_parser(
        "generate",
        help="Create or update VHDL source files based on the directives within existing files.",
    )
    generate_parser.add_argument("path", help=path_help)

    return parser.parse_args()


def get_files_to_work_on(path):
    if not path.startswith("/"):
        path = os.getcwd() + "/" + path

    files = []
    if os.path.isfile(path):
        files.append(path)
    elif os.path.isdir(path):
        for f in pathlib.Path(path).glob("**/*.vhd"):
            files.append(f)
    else:
        raise Exception(f"{args.path} is not a path to file or directory")

    return files


def main():
    args = parse_cmd_line_args()

    files = get_files_to_work_on(args.path)

    sys.exit(args.func(files))


if __name__ == "__main__":
    main()



import unittest
import io

from thdl.check.reset import check

LINES_STUCK_TO_0 = """rst_n => '0',
rst_n=>'0',
rstn => '0'
reset_n_i=>'0',
reset_n => '0',
reset_n_i => '0',
RST_N_I=>'0');
wb_rst_n=> '0',
wb_rst_n_i=> '0',
foo_bar_reset_n => '0',
foo_rstn=>'0',
"""

LINES_MAPPED_TO_POSITIVE = """rst_n => rst,
wb_reset_n => reset,
rstn => resetp,
wb_rst_n => reset);
foo_rst_n_i => rst_i,
"""

LINES_NEGATED_NEGATIVE = """rst_n => not rst_n_i,
resetn => not(rst_n),
rst_i_n => not wb_resetn,
reset_n_i => not rstn);
"""

LINES_VALID ="""rst_n => '1',
rst_n_i => not rst_p_i,
rst_n_i => not(reset_p),
resetn => not wb_rst_i
rst_i_n => not wb_reset,
"""

class TestPositiveReset(unittest.TestCase):
    def test_stuck_to_0(self):
        fh = io.StringIO(LINES_STUCK_TO_0)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, "Negative reset stuck to '0'!", l)

    def test_mapped_to_positive(self):
        fh = io.StringIO(LINES_MAPPED_TO_POSITIVE)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, "Negative reset mapped to positive reset!", l)

    def test_negated_negative(self):
        fh = io.StringIO(LINES_NEGATED_NEGATIVE)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, "Negative reset mapped to negated negative reset!", l)

    def test_valid(self):
        fh = io.StringIO(LINES_VALID)

        for l in fh:
            msg = check(l.lower(), silent=True)
            self.assertEqual(msg, None, l)



import re

from . import record
from . import enumeration

def generate(filepath):
    print("Hello from generate!")
    record.record()

    with open(filepath) as fh:
        _parse_file(fh)


def _parse_file(fh):

    # Indicates whether  "--thdl:generate" tag has been found.
    # It is reset on empty lines and after generations.
    tag_found = False

    for line in fh:
        line = line.strip()

        if line.startswith("--thdl:generate"):
            tag_found = True
            continue
        elif tag_found == False:
            continue

        if not line:
            tag_found = False
            continue

        tag_found = False
        if re.match("type\s+\w+\s+is\s*\(", line, re.I):
            enumeration.Enumeration(fh, line)
#        elif:
        else:
            tag_found = True



import re

from . import file_info

POSITIVE_RESET_PORT_MAP = re.compile("re?se?t((p)|(_p)|(_i)|(_p_i)|(_i_p))?\s*=>")
POSITIVE_RESET = re.compile("re?se?t((p)|(_i)|(_p))?")

NEGATIVE_RESET_PORT_MAP = re.compile("re?se?t((n)|(_n)|(_n_i)|(_i_n))\s*=>")
NEGATIVE_RESET = re.compile("re?se?t((n)|(_n)|(_i_n))")

STARTS_WITH_NOT = re.compile("^not\s*\(?")


SILENT = False


def check(line, silent=False):
    """Check line for stupid reset mistakes.

    Parameters:
    -----------
    line :
        Line read from file.
    silent : bool
        Do not print any message, only return it.
        Useful for unit tests.

    Returns
    -------
        Reference to string if violation is found. Otherwise None.
    """
    global SILENT
    SILENT = silent

    line = line.strip()
    if line.startswith("--"):
        return None

    if POSITIVE_RESET_PORT_MAP.search(line):
        return _positive_reset(line)

    if NEGATIVE_RESET_PORT_MAP.search(line):
        return _negative_reset(line)


def _message(msg):
    if not SILENT:
        print("{}:{}".format(file_info.FILEPATH, file_info.LINE_NUMBER))
        print(file_info.LINE, end='')
        print(msg + "\n")

    return msg


def _positive_reset(line):
    assignee = line.split("=>")[1].strip()

    if assignee.startswith("'1'"):
        return _message("Positive reset stuck to '1'!")

    negated = False
    if STARTS_WITH_NOT.search(assignee):
        negated = True

    if NEGATIVE_RESET.search(assignee):
        reset = "negative"
    elif POSITIVE_RESET.search(assignee):
        reset = "positive"
    else:
        return None

    if reset == "negative" and not negated:
        return _message("Positive reset mapped to negative reset!")

    if reset == "positive" and negated:
        return _message("Positive reset mapped to negated positive reset!")


def _negative_reset(line):
    assignee = line.split("=>")[1].strip()

    if assignee.startswith("'0'"):
        return _message("Negative reset stuck to '0'!")

    negated = False
    if STARTS_WITH_NOT.search(assignee):
        negated = True

    if NEGATIVE_RESET.search(assignee):
        reset = "negative"
    elif POSITIVE_RESET.search(assignee):
        reset = "positive"
    else:
        return None

    if reset == "positive" and not negated:
        return _message("Negative reset mapped to positive reset!")

    if reset == "negative" and negated:
        return _message("Negative reset mapped to negated negative reset!")
